"""A template of the config json file named "api_scopus_config.json" is stored
in the ".ScopusApyJson/CONFIG" folder. The user's "api_scopus_config.json"
should be located in the user's folder returned by the function `_get_config_dir`
and the user is invited to fill this file with its scopus api authentication keys.

If the user's "api_scopus_config.json" file exits in this user's folder,
it will be used.
Otherwise, a copy of the template file of the ".ScopusApyJson/CONFIG"
folder of the ScopusApyJson package will be automatically created in the user's
folder in order to be filled with the user's scopus api authentication keys.
"""

__all__ = ['API_CONFIG_DICT',
           'API_CONFIG_PATH',
           'ELS_LINK',
           'PARSED_SCOPUS_COLUMNS_NAMES',
           'SELECTED_SCOPUS_COLUMNS_NAMES',
          ]

# Standard library imports
import json
import os.path
import sys
import warnings
from pathlib import Path


# Header of the query for the Scopus API
ELS_LINK = "https://api.elsevier.com/content/abstract/"


# Internal functions to build the Globals
def _get_config_dir():
    """
    Returns a parent directory path
    where persistent application data can be stored.

    # linux: ~/.local/share
    # macOS: ~/Library/Application Support
    # windows: C:/Users/<USER>/AppData/Roaming
    adapted from : ("https://stackoverflow.com/questions/19078969/"
    "python-getting-appdata-folder-in-a-cross-platform-way")
    """
    home = Path.home()

    if sys.platform=='win32':
        app_data_path = home / Path('AppData/Roaming')
    elif sys.platform=='darwin':
        app_data_path = home / Path('Library/Application Support')
    else:
        app_data_path = home / Path('.local/share')
    return app_data_path


def _dump_json(file_path, json_dict):
    with open(file_path, 'w', encoding="utf-8") as file:
        json.dump(json_dict, file, indent=4)


def _check_api_keys(config_path, config_dict):
    dict_apikey = config_dict["apikey"]
    dict_insttoken = config_dict["insttoken"]
    apikey_undef = dict_apikey in  ["PAST_APIKEY_HERE", ""]
    insttoken_undef = dict_insttoken in ["PAST_INSTTOKEN_HERE", ""]
    if apikey_undef or insttoken_undef:
        message = "Authentication keys are not yet defined (required).\n"
        warnings.warn(message)
        config_dict["apikey"] = input(("Enter your authentication key "
                                       "(obtained from http://dev.elsevier.com):"))
        config_dict["insttoken"] = input(("your institution token provided "
                                          "by Elsevier support staff:"))
        _dump_json(config_path, config_dict)
    return config_dict


def _config_saj_dict(json_file_name):
    # Reads the default json_file_name config file
    pck_config_path = Path(__file__).parent / Path('CONFIG') / Path(json_file_name)
    pck_config_date = os.path.getmtime(pck_config_path)
    with open(pck_config_path, encoding="utf-8") as file:
        json_config_dict = json.load(file)

    # Sets the json_config_dict according to the status of the local config file
    local_config_dir_path  = _get_config_dir() / Path('ScopusApyJson')
    local_config_file_path = local_config_dir_path  / Path(json_file_name)

    if os.path.exists(local_config_file_path):
        # Local json_file_name config file exists
        local_config_date = os.path.getmtime(local_config_file_path)
        if local_config_date > pck_config_date:
            # Local config file is more recent than package one
            # thus json_config_dict is defined by the local config file
            with open(local_config_file_path, encoding="utf-8") as file:
                json_config_dict = json.load(file)
        else:
            # Local config file is less recent than package one
            # thus package config file overwrite the local config file
            # and json_config_dict is kept at default values
            _dump_json(local_config_file_path, json_config_dict)
    else:
        # Local json_file_name config file does not exist
        # thus package config file is used to create a local config file
        # to be filled by the user
        # and json_config_dict is kept at default values
        if not os.path.exists(_get_config_dir() / Path('ScopusApyJson')):
            os.makedirs(_get_config_dir() / Path('ScopusApyJson'))
        _dump_json(local_config_file_path, json_config_dict)

    return json_config_dict, local_config_file_path

# Getting the authentication key and the institution token
# through the json file 'api_scopus_config.json'
# stored in the folder ".ScopusApyJson/CONFIG"
config_dict, config_path = _config_saj_dict('api_scopus_config.json')
API_CONFIG_DICT = _check_api_keys(config_path, config_dict)
API_CONFIG_PATH = config_path

# Getting the names of the selected scopus columns by the user
# through the json file "scopus_col_names.json"
# stored in the folder ""~/AppData/Roaming/ScopusApyJson" of the user
scopus_column_names_dict, _ = _config_saj_dict('scopus_col_names.json')
PARSED_SCOPUS_COLUMNS_NAMES = list(scopus_column_names_dict.keys())[1:]
SELECTED_SCOPUS_COLUMNS_NAMES = [k for k,v in scopus_column_names_dict.items() if v][1:]



"""Functions for building Scopus data through the API based on DOI."""

__all__ = ['build_scopus_df_from_api',]

# 3rd party imports
import pandas as pd

# Local imports
import ScopusApyJson.saj_globals as saj_g
from ScopusApyJson.api_manager import get_doi_json_data_from_api
from ScopusApyJson.json_parser import parse_json_data_to_scopus_df


def build_scopus_df_from_api(doi_list, timeout = None, verbose = False):
    """The function `build_scopus_df_from_api` gets, for each of the DOI 
    in the DOI list "doi_list", the hierarchical dict of the data returned by 
    the function `get_doi_json_data_from_api`.

    Then, it builts the dataframes resulting from the hierarchical dicts parsing 
    using the function `parse_json_data_to_scopus_df`. 
    Finally, it concatenates these dataframe to a single dataframe.

    Args:
        doi_list (list): The list of DOI (str) for the Scopus API request.
        timeout (int): The maximum waiting time in seconds for request answer.
    Returns:
        ((pandas.core.frame.DataFrame, pandas.core.frame.DataFrame, bool)): Tuple where first term \
        is the concatenation of the dataframes resulting from the hierarchical dicts parsing \
        for each DOI of the list "doi_list", second term is the dataframe giving the failed DOIs \
        with the reasons of their fail and third term is equal to False if the authentication \
        to the scopus database failed.
    """
    if not isinstance(doi_list, list):
        doi_list = [doi_list]
    scopus_df_list = []
    failed_num = 0
    authy_status = False
    failed_doi_df = pd.DataFrame(columns = ["DOI","Fail reason"])
    api_scopus_df = pd.DataFrame(columns = saj_g.SELECTED_SCOPUS_COLUMNS_NAMES)
    for _, doi in enumerate(doi_list):
        api_json_data, request_status = get_doi_json_data_from_api(doi, timeout = timeout)
        if request_status == "False":
            authy_status = False
            if verbose:
                print(('Authentication failed: please check availability'
                       ' of authentication keys'))
            break
        authy_status = True
        if request_status == "Empty":
            if verbose:
                print(f'DOI {doi} not found in Scopus database')
            failed_doi_df.loc[failed_num, "DOI"] = doi
            failed_doi_df.loc[failed_num, "Fail reason"] = "Not found"
            failed_num += 1
        elif request_status == "Timeout":
            if verbose:
                print('Request timeout: please check web access')
            failed_doi_df.loc[failed_num, "DOI"] = doi
            failed_doi_df.loc[failed_num, "Fail reason"] = "Timeout"
            failed_num += 1
        else:
            if verbose:
                print(f'Request successful for DOI {doi}')
            scopus_df = parse_json_data_to_scopus_df(api_json_data)
            scopus_df_list.append(scopus_df)
            api_scopus_df = pd.concat(scopus_df_list, axis = 0)

    return api_scopus_df, failed_doi_df, authy_status



"""Functions for Scopus API management with request based on DOI."""

__all__ = ['get_doi_json_data_from_api',]

# Standard library imports
import json

# 3rd party imports
import requests
from requests.exceptions import Timeout

# Internal imports
import ScopusApyJson.saj_globals as saj_g


def _set_els_doi_api(MyScopusKey, MyInstKey, doi):
    """The internal function `_set_els_doi_api` sets, for the DOI 'doi', 
    the query 'els_api' according to the Scopus API usage 
    which header is given by the global 'ELS_LINK'.
    
    Args:
        MyScopusKey (str): The user's authentication key.
        MyInstKey (str): The user's institution token.
        doi (str): The publication DOI for which the Scopus API \
        will provide information. 
    Returns:
        (str): The query for the passed DOI according to the scopus \
        api usage.
    """
    # Setting the query
    query_header = saj_g.ELS_LINK
    query = doi + '?'

    # Building the HAL API
    els_api = query_header \
            + query \
            + '&apikey='    + MyScopusKey \
            + '&insttoken=' + MyInstKey \
            + '&httpAccept=application/json'

    return els_api


def _get_json_from_api(doi, api_config_dict, timeout):
    """The internal function `_get_json_from_api` gets, for the DOI 'doi', 
    the response to the query 'els_api' built using the `_set_els_doi_api` 
    internal function .

    It passes to this function the user's authentication key 'MyScopusKey' 
    and the user's institutional token 'MyInstKey' given by the 'api_config_dict' 
    dict. 
    It also increments the number of requests performed by the user. The number 
    is updated in the dict 'api_config_dict' at key 'api_uses_nb'.

    Args:
        doi (str): The publication DOI for which the Scopus API will provide data.
        api_config_dict (dict): The dict wich values are the user's authentication \
        key, the user's institutional token and the number of requests performed.
        timeout (int): The maximum waiting time in seconds for request answer.
    Returns:
        (tup): The tup composed by the hierarchical-dict response \
        to the query and the updated 'api_config_dict' dict.
    """
    # Initializing parameters
    response_dict = None
    response = None
    response_status = "False"

    # Setting client authentication keys
    MyScopusKey = api_config_dict["apikey"]
    MyInstKey   = api_config_dict["insttoken"]
    api_uses_nb = api_config_dict['api_uses_nb']

    # Setting Elsevier API
    els_api = _set_els_doi_api(MyScopusKey, MyInstKey, doi)

    # Get the request response
    try:
        response = requests.get(els_api, timeout = timeout)
    except Timeout:
        response_status = "Timeout"
    else:
        if any(x==response.status_code for x in [204, 404]):
            response_status = "Empty"
        else:
            response_status = "True"
            response_dict = response.json()
        # Updating api_uses_nb in config_dict
        api_config_dict["api_uses_nb"] = api_uses_nb + 1

    return response_dict, api_config_dict, response_status


def _update_api_config_json(api_config_path, api_config_dict):
    with open(api_config_path, 'w', encoding="utf-8") as f:
        json.dump(api_config_dict, f, indent = 4)


def get_doi_json_data_from_api(doi, timeout = None):
    """The function `get_doi_json_data_from_api` gets, for the DOI 'doi', 
    the json-serialized response to the Scopus API request using 
    the internal function `_get_json_from_api`.
    It passes to this function the user's dict 'API_CONFIG_DICT'. 
    It also updates the API configuration json file with the modified 
    dict 'API_CONFIG_DICT' returned by this function.
    
    Args:
        doi (str): The publication DOI for which the Scopus API \
        will provide data.
        timeout (int): The maximum waiting time in seconds \
        for request answer (default: 10).
    Returns:
        (dict): The hierarchical dict of the data returned \
        by the internal function '_get_json_from_api'.
    """
    if not timeout:
        timeout = 10

    # Getting api json data
    return_tup = _get_json_from_api(doi, saj_g.API_CONFIG_DICT,
                                    timeout)
    doi_json_data, api_config_dict, request_status = return_tup

    # Updatting api config json with number of requests
    _update_api_config_json(saj_g.API_CONFIG_PATH,
                            api_config_dict)

    return doi_json_data, request_status
